﻿using System;
using System.IdentityModel.Services;
using System.IdentityModel.Tokens;
using System.IO;
using System.Linq;
using System.Security.Claims;
using System.ServiceModel;
using System.Threading;
using System.Xml;
using DotNetSamples.Common;
using DotNetSamples.Common.Contract;
using DotNetSamples.Common.JavaServiceSample;

namespace DotNetSamples.Service
{
    /// <summary>
    /// Implementering af servicekontrakt
    /// </summary>
    public class ServiceImpl : IServiceContract
    {
        /// <summary>
        /// Hello world operation
        /// </summary>
        public HelloWorldResult HelloWorld(HelloWorldQuery query)
        {
            // Verificer, at claims kan ses
            var identity = GetClaimsIdentity();

            return new HelloWorldResult
                       {
                           Text = "Hello .NET!",
                           ClaimsCount = identity.Claims.Count()
                       };
        }

        /// <summary>
        /// Operation der verificerer at bootstrap token'et er tilgængeligt
        /// </summary>
        public void AssertBootstrapTokenIsAvailable()
        {
            var identity = GetClaimsIdentity();

            var token = identity.BootstrapContext as BootstrapContext;
            if (EnsureBootstrapSecurityToken(token) == null)
                throw new ApplicationException("Cannot get boostrap context from current identity.");

            if (token == null)
                throw new ApplicationException("Bootstrap token cannot be null");
        }

        public JavaServiceResult CallJavaServiceWithActAs()
        {
            // Verificer, at der et bootstrap token
            AssertBootstrapTokenIsAvailable();

            // Tag logon principal
 
            var identity = Thread.CurrentPrincipal.Identity as ClaimsIdentity;
            if (identity == null) throw new ApplicationException("ClaimsPrincipal must exist");
            // Hiv fat i token'et
            var token = identity.BootstrapContext as BootstrapContext;
            if (token == null)
                throw new ApplicationException("Cannot get boostrap context from current identity.");


            // Request nyt token
            var newToken = WsTrustClient.RequestSecurityTokenWithUserName(
                Constants.StsAddressUserName,
                Constants.StsCertificate,
                Constants.JavaServiceAddress,
                Constants.DmpUserName,
                Constants.DmpPassword,
                EnsureBootstrapSecurityToken(token));

            // Kald Java service
            var javaResult = CallJavaService(newToken);

            // Returner til kalder
            return new JavaServiceResult
                       {
                           Text = javaResult
                       };
        }

        private static SecurityToken EnsureBootstrapSecurityToken(BootstrapContext bootstrapContext)
        {
            if (bootstrapContext.SecurityToken != null)
                return bootstrapContext.SecurityToken;
            if (string.IsNullOrWhiteSpace(bootstrapContext.Token))
                return null;
            var handlers = FederatedAuthentication.FederationConfiguration.IdentityConfiguration.SecurityTokenHandlers;
            return handlers.ReadToken(new XmlTextReader(new StringReader(bootstrapContext.Token)));
        }

        private static string CallJavaService(SecurityToken securityToken)
        {
            // Lav binding til STS'en
            var stsBinding = WsTrustClient.GetStsBinding();

            // Lav binding til servicen
            var serviceBinding = WsTrustClient.GetServiceBinding(stsBinding, Constants.StsAddressUserName, false);

            // Deaktiver secure conversation, da Java servicen ikke bruger det
            var serviceBindingWithoutSecureConversation = WsTrustClient.CreateWsFederationBindingWithoutSecureConversation(serviceBinding);

            // Lav channel til servicen
            var channelFactory = new ChannelFactory<JavaSamples>(serviceBindingWithoutSecureConversation, Constants.JavaServiceAddress);

            if (channelFactory.Credentials == null)
                throw new ApplicationException("ChannelFactory must have credentials");

            // Sæt encryption certifikat til java servicen
            channelFactory.Credentials.ServiceCertificate.ScopedCertificates.Add(Constants.JavaServiceAddress.Uri, Constants.ServiceCertificate);

            // Lav channel med token'et fra STS'en
            var channel = channelFactory.CreateChannelWithIssuedToken(securityToken);
            try
            {
                // Kald service
                var response = channel.HelloWorld(new HelloWorldRequest());

                if(response == null)
                    throw new ApplicationException("Java response was null");

                if(response.Body == null)
                    throw new ApplicationException("Java body was null");

                // Returner
                return response.Body.@return;
            }
            finally
            {
                // Luk channel
                WsTrustClient.CloseChannel(channel);
            }
        }

        /// <summary>
        /// Hiver claims identity ud for brugeren der er logget på
        /// </summary>
        private static ClaimsIdentity GetClaimsIdentity()
        {
            if (OperationContext.Current != null)
            {
                Thread.CurrentPrincipal = OperationContext.Current.ClaimsPrincipal;
            }

            var claimsPrincipal = Thread.CurrentPrincipal as ClaimsPrincipal;

            if (claimsPrincipal == null)
                throw new ApplicationException("ClaimsPrincipal must exist");

            if (claimsPrincipal.Identities.Count() < 1)
                throw new ApplicationException("IClaimsIdentityt must exist");

            var identity = claimsPrincipal.Identities.First();

            return identity;
        }
    }
}