﻿using System;
using System.IdentityModel.Tokens;
using System.Security.Cryptography.X509Certificates;
using System.ServiceModel;
using System.ServiceModel.Security;
using DotNetSamples.Common;
using DotNetSamples.Common.Contract;
using DotNetSamples.Common.JavaServiceSample;

namespace DotNetSamples.Tests
{
    public static class TestCommon
    {
        public static HelloWorldResult TestCanCallServiceWithIssuedToken(
            SecurityToken securityToken)
        {
            // Lav binding til STS'en
            var stsBinding = WsTrustClient.GetStsBinding();

            // Lav binding til STS servicen
            var serviceBinding = WsTrustClient.GetServiceBinding(stsBinding, Constants.StsAddressUserName, true);

            // Lav channel til DotNet servicen
            var channelFactory = new ChannelFactory<IServiceContract>(serviceBinding, Constants.DotNetServiceAddress);

            if(channelFactory.Credentials == null)
                throw new ApplicationException("ChannelFactory must have credentials");

            // Deaktiver service cert validering
            channelFactory.Credentials.ServiceCertificate.Authentication.CertificateValidationMode = X509CertificateValidationMode.None;
            channelFactory.Credentials.ServiceCertificate.Authentication.RevocationMode = X509RevocationMode.NoCheck;
            channelFactory.Credentials.UseIdentityConfiguration = true;

            // Påhæft token til kald til service
            var channel = channelFactory.CreateChannelWithIssuedToken(securityToken);
            try
            {
                // Lav query
                var query = new HelloWorldQuery
                {
                    Text = "Jesper Hvid"
                };

                // Kald service
                return channel.HelloWorld(query);
            }
            finally
            {
                // Luk channel
                WsTrustClient.CloseChannel(channel);
            }
        }

        public static void TestCanCallServiceWithIssuedTokenAndVerifyBootstrap(
            SecurityToken securityToken)
        {
            // Lav binding til STS'en
            var stsBinding = WsTrustClient.GetStsBinding();

            // Lav binding til servicen
            var serviceBinding = WsTrustClient.GetServiceBinding(stsBinding, Constants.StsAddressUserName, true);

            // Lav channel factory til service
            var channelFactory = new ChannelFactory<IServiceContract>(serviceBinding, Constants.DotNetServiceAddress);

            if (channelFactory.Credentials == null)
                throw new ApplicationException("ChannelFactory must have credentials");

            // Deaktiver service cert validering
            channelFactory.Credentials.ServiceCertificate.Authentication.CertificateValidationMode =
                X509CertificateValidationMode.None;
            channelFactory.Credentials.ServiceCertificate.Authentication.RevocationMode = X509RevocationMode.NoCheck;


            // Påhæft token til kald til service
            var channel = channelFactory.CreateChannelWithIssuedToken(securityToken);
            try
            {
                // Kald service
                channel.AssertBootstrapTokenIsAvailable();
            }
            finally
            {
                // Luk channel
                WsTrustClient.CloseChannel(channel);
            }
        }

        public static JavaServiceResult TestCanCallServiceWithIssuedTokenAndVerifyBootstrapAndCallJavaServiceWithActAs(
            SecurityToken securityToken)
        {
            // Lav binding til STS'en
            var stsBinding = WsTrustClient.GetStsBinding();

            // Lav binding til servicen
            var serviceBinding = WsTrustClient.GetServiceBinding(stsBinding, Constants.StsAddressUserName, true);

            // Lav channel factory til service
            var channelFactory = new ChannelFactory<IServiceContract>(serviceBinding, Constants.DotNetServiceAddress);

            if (channelFactory.Credentials == null)
                throw new ApplicationException("ChannelFactory must have credentials");

            // Deaktiver service cert validering
            channelFactory.Credentials.ServiceCertificate.Authentication.CertificateValidationMode =
                X509CertificateValidationMode.None;
            channelFactory.Credentials.ServiceCertificate.Authentication.RevocationMode = X509RevocationMode.NoCheck;


            // Påhæft token til kald til service
            var channel = channelFactory.CreateChannelWithIssuedToken(securityToken);
            try
            {
                // Kald service
                return channel.CallJavaServiceWithActAs();
            }
            finally
            {
                // Luk channel
                WsTrustClient.CloseChannel(channel);
            }
        }

        public static HelloWorldResponse TestCanCallJavaServiceWithIssuedToken(
            SecurityToken securityToken)
        {
            // Lav binding til STS'en
            var stsBinding = WsTrustClient.GetStsBinding();

            // Lav binding til servicen
            var serviceBinding = WsTrustClient.GetServiceBinding(stsBinding, Constants.StsAddressUserName, false);

            // Deaktiver secure conversation, da Java servicen ikke bruger det
            var serviceBindingWithoutSecureConversation = WsTrustClient.CreateWsFederationBindingWithoutSecureConversation(serviceBinding);

            // Lav channel til servicen
            var channelFactory =
                new ChannelFactory<JavaSamples>(
                    serviceBindingWithoutSecureConversation,
                    Constants.JavaServiceAddress);

            if (channelFactory.Credentials == null)
                throw new ApplicationException("ChannelFactory must have credentials");

            // Sæt encryption certifikat til java servicen
            channelFactory.Credentials.ServiceCertificate.ScopedCertificates.Add(Constants.JavaServiceAddress.Uri, Constants.ServiceCertificate);

            // Lav channel med token'et fra STS'en
            var channel = channelFactory.CreateChannelWithIssuedToken(securityToken);
            try
            {
                // Lav query
                //var query = new DotNetSamples.Common.DAI.
                //                {
                //                   Body = new Common.Tomcat.helloRequestBody("Jesper")
                //                };

                // Kald service
                return channel.HelloWorld(new HelloWorldRequest());
            }
            finally
            {
                // Luk channel
                WsTrustClient.CloseChannel(channel);
            }
        }
    }
}
